from ._distributions import Distribution, DistributionCollector, get_distributions, get_distributions_of
from ._virtualenv import VirtualEnvInfo, get_current_venv

__all__ = [
    "Distribution",
    "DistributionCollector",
    "get_distributions",
    "get_distributions_of",
    "VirtualEnvInfo",
    "get_current_venv",
]



""" This module provides provides the :class:`Supplier` interface which is used to represent values that can be
calculated lazily and track provenance of such computations. """


import abc
from typing import Any, Callable, Generic, Iterable, List, Sequence, Tuple, TypeVar

from ._generic import NotSet

T = TypeVar("T", covariant=True)
U = TypeVar("U")


class Supplier(Generic[T], abc.ABC):
    """Base class for value suppliers."""

    class Empty(Exception):
        """Raised when a supplier cannot provide a value."""

        def __init__(self, supplier: "Supplier[Any]", message: "str | None" = None) -> None:
            self.supplier = supplier
            self.message = message

        def __str__(self) -> str:
            if self.message:
                return f"{self.message} ({self.supplier})"
            else:
                return str(self.supplier)

    @abc.abstractmethod
    def derived_from(self) -> Iterable["Supplier[Any]"]:
        """Return an iterable that yields all suppliers that this supplier is derived from."""

    @abc.abstractmethod
    def get(self) -> T:
        """Return the value of the supplier. Depending on the implemenmtation, this may defer to other suppliers."""

    def get_or(self, fallback: U) -> "T | U":
        """Return the value of the supplier, or the *fallback* value if the supplier is empty."""
        try:
            return self.get()
        except Supplier.Empty:
            return fallback

    def get_or_raise(self, get_exception: Callable[[], Exception]) -> T:
        """Return the value of the supplier, or raise the exception provided by *get_exception* if empty."""
        try:
            return self.get()
        except Supplier.Empty:
            raise get_exception()

    def is_empty(self) -> bool:
        """Returns `True` if the supplier is empty."""
        try:
            self.get()
        except Supplier.Empty:
            return True
        else:
            return False

    def is_filled(self) -> bool:
        """Returns `True` if the supplier is not empty."""
        return not self.is_empty()

    def is_void(self) -> bool:
        return False

    def map(self, func: Callable[[T], U]) -> "Supplier[U]":
        """Maps *func* over the value in the supplier."""

        return MapSupplier(func, self)

    def once(self) -> "Supplier[T]":
        """Cache the value forever once :attr:`get` is called."""

        return OnceSupplier(self)

    def lineage(self) -> Iterable[Tuple["Supplier[Any]", List["Supplier[Any]"]]]:
        """Iterates over all suppliers in the lineage.

        Yields:
            A supplier and the suppliers it is derived from.
        """

        stack: List["Supplier[Any]"] = [self]
        while stack:
            current = stack.pop(0)
            derived_from = list(current.derived_from())
            yield current, derived_from
            stack += derived_from

    @staticmethod
    def of(value: "T | Supplier[T]", derived_from: Sequence["Supplier[Any]"] = ()) -> "Supplier[T]":
        if isinstance(value, Supplier):
            return value
        return OfSupplier(value, derived_from)

    @staticmethod
    def of_callable(func: Callable[[], T], derived_from: Sequence["Supplier[Any]"] = ()) -> "Supplier[T]":
        return OfCallableSupplier(func, derived_from)

    @staticmethod
    def void(from_exc: "Exception | None" = None, derived_from: Sequence["Supplier[Any]"] = ()) -> "Supplier[T]":
        """Returns a supplier that always raises :class:`Empty`."""

        return VoidSupplier(from_exc, derived_from)

    def __repr__(self) -> str:
        try:
            value = self.get_or(NotSet.Value)
        except Exception as exc:
            inner = f"<exception reading value: {exc}>"
        else:
            if value is NotSet.Value:
                inner = "<empty>"
            else:
                inner = f"value={value!r}"
        return f"{type(self).__name__}({inner})"


class MapSupplier(Supplier[U], Generic[T, U]):
    def __init__(self, func: Callable[[T], U], value: Supplier[T]) -> None:
        self._func = func
        self._value = value

    def derived_from(self) -> Iterable[Supplier[Any]]:
        yield self._value

    def get(self) -> U:
        try:
            return self._func(self._value.get())
        except Supplier.Empty:
            raise Supplier.Empty(self)

    def __repr__(self) -> str:
        return f"{self._value}.map({self._func})"

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        assert isinstance(other, MapSupplier)
        return (self._func, self._value) == (other._func, other._value)


class OnceSupplier(Supplier[T]):
    _value: "T | NotSet" = NotSet.Value
    _empty: "Supplier.Empty | None" = None

    def __init__(self, delegate: Supplier[T]) -> None:
        self._delegate = delegate

    def derived_from(self) -> Iterable[Supplier[Any]]:
        yield self._delegate

    def get(self) -> T:
        if self._empty is not None:
            raise Supplier.Empty(self) from self._empty
        if self._value is NotSet.Value:
            try:
                self._value = self._delegate.get()
            except Supplier.Empty as exc:
                self._empty = exc
                raise Supplier.Empty(self) from exc
        return self._value

    def __repr__(self) -> str:
        return f"{self._delegate}.once()"

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        assert isinstance(other, OnceSupplier)
        return (self._delegate,) == (other._delegate,)


class OfCallableSupplier(Supplier[T]):
    def __init__(self, func: Callable[[], T], derived_from: Sequence[Supplier[Any]]) -> None:
        self._func = func
        self._derived_from = derived_from

    def derived_from(self) -> Iterable[Supplier[Any]]:
        return self._derived_from

    def get(self) -> T:
        return self._func()

    def __repr__(self) -> str:
        return f"Supplier.of_callable({self._func})"

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        assert isinstance(other, OfCallableSupplier)
        return (self._func, self._derived_from) == (other._func, other._derived_from)


class OfSupplier(Supplier[T]):
    def __init__(self, value: T, derived_from: Sequence[Supplier[Any]]) -> None:
        self._value = value
        self._derived_from = tuple(derived_from)

    def derived_from(self) -> Iterable[Supplier[Any]]:
        return self._derived_from

    def get(self) -> T:
        return self._value

    def __repr__(self) -> str:
        return f"Supplier.of({self._value})"

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        assert isinstance(other, OfSupplier)
        return (self._value, self._derived_from) == (other._value, other._derived_from)


class VoidSupplier(Supplier[T]):
    def __init__(self, from_exc: "Exception | None", derived_from: Sequence[Supplier[Any]]) -> None:
        self._from_exc = from_exc
        self._derived_from = tuple(derived_from)

    def derived_from(self) -> Iterable[Supplier[Any]]:
        return self._derived_from

    def get(self) -> T:
        raise Supplier.Empty(self) from self._from_exc

    def is_void(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"Supplier.void(from_exc={self._from_exc})"

    def __hash__(self) -> int:
        return hash((self._from_exc, self._derived_from))

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        assert isinstance(other, VoidSupplier)
        return (self._from_exc, self._derived_from) == (other._from_exc, other._derived_from)



"""
Implements build script runners.
"""

import re
import types
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Iterable, NamedTuple, Sequence

import builddsl

from ._buildscript import BuildscriptMetadata, buildscript
from ._generic import NotSet

##
# Data classes
##


class ProjectInfo(NamedTuple):
    script: Path
    runner: "ScriptRunner"


##
# Interfaces
##


class ScriptRunner(ABC):
    """
    Abstract class for script runners. Implementations of this class are used to detect a script in a directory
    and to actually run it. The Kraken wrapper and build system both use this to run a build script, which for
    the wrapper is needed to extract the build script metadata.
    """

    @abstractmethod
    def find_script(self, directory: Path) -> "Path | None":
        raise NotImplementedError(self)

    @abstractmethod
    def execute_script(self, script: Path, scope: Dict[str, Any]) -> None:
        raise NotImplementedError(self)

    @abstractmethod
    def has_buildscript_call(self, script: Path) -> bool:
        """
        Implement a heuristic to check if the script implements a call to the :func:`buildscript` function.
        """

        raise NotImplementedError(self)

    @abstractmethod
    def get_buildscript_call_recommendation(self, metadata: BuildscriptMetadata) -> str:
        """
        Make a recommendation to the user for the code the user should put into their build script for the
        :func:`buildscript` call that is required by Kraken wrapper.
        """

        raise NotImplementedError(self)


class ProjectFinder(ABC):
    """
    Base class for finding a Kraken project starting from any directory.
    """

    @abstractmethod
    def find_project(self, directory: Path) -> "ProjectInfo | None":
        ...


##
# ScriptRunner Implementations
##


class ScriptPicker(ScriptRunner):
    """
    Base class for picking the right script file in a directory based on a few criteria.
    """

    def __init__(self, filenames: Sequence[str]) -> None:
        self.filenames = list(filenames)

    def find_script(self, directory: Path) -> "Path | None":
        for filename in self.filenames:
            script = directory / filename
            if script.is_file():
                return script
        return None


class PythonScriptRunner(ScriptPicker):
    """
    A finder and runner for Python based Kraken build scripts called `.kraken.py`.

    !!! note

        We can't call the script `kraken.py` (without the leading dot), as otherwise under most circumstances the
        script will try to import itself when doing `import kraken` or `from kraken import ...`.
    """

    def __init__(self, filenames: Sequence[str] = (".kraken.py",)) -> None:
        super().__init__(filenames)

    def execute_script(self, script: Path, scope: Dict[str, Any]) -> None:
        module = types.ModuleType(str(script.parent))
        module.__file__ = str(script)

        code = compile(script.read_text(), script, "exec")
        exec(code, vars(module))

    def has_buildscript_call(self, script: Path) -> bool:
        code = script.read_text()
        if not re.search(r"^from kraken.common import buildscript", code, re.M):
            return False
        if not re.search(r"^buildscript\s*\(", code, re.M):
            return False
        return True

    def get_buildscript_call_recommendation(self, metadata: BuildscriptMetadata) -> str:
        code = "from kraken.common import buildscript\nbuildscript("
        if metadata.index_url:
            code += f"\n    index_url={metadata.index_url!r},"
        if metadata.extra_index_urls:
            if len(metadata.extra_index_urls) == 1:
                code += f"\n    extra_index_urls={metadata.extra_index_urls!r},"
            else:
                code += "\n    extra_index_urls=["
                for url in metadata.extra_index_urls:
                    code += f"\n        {url!r},"
                code += "\n    ],"
        if metadata.requirements:
            if sum(map(len, metadata.requirements)) < 50:
                code += f"\n    requirements={metadata.requirements!r},"
            else:
                code += "\n    requirements=["
                for req in metadata.requirements:
                    code += f"\n        {req!r},"
                code += "\n    ],"
        if not code.endswith("("):
            code += "\n"
        code += ")"
        return code


class BuildDslScriptRunner(ScriptPicker):
    """
    A finder and runner for BuildDSL based Kraken build scripts called `kraken.build` (optionally prefixed with `.`).
    """

    def __init__(self, filenames: Sequence[str] = ("kraken.build", ".kraken.build")) -> None:
        super().__init__(filenames)

    def execute_script(self, script: Path, scope: Dict[str, Any]) -> None:
        code = script.read_text()
        scope = {"buildscript": buildscript, **scope}
        builddsl.Closure.from_map(scope).run_code(code, str(script))

    def has_buildscript_call(self, script: Path) -> bool:
        code = script.read_text()
        if re.match(r"^buildscript\s*(\(|\{})", code, re.M):
            return False
        return True

    def get_buildscript_call_recommendation(self, metadata: BuildscriptMetadata) -> str:
        code = "buildscript {"
        if metadata.index_url:
            code += f"\n    index_url = {metadata.index_url!r}"
        if metadata.extra_index_urls:
            for url in metadata.extra_index_urls:
                code += f"\n    extra_index_url {url!r}"
        if metadata.requirements:
            for req in metadata.requirements:
                code += f"        requires {req!r}"
        code += "}"
        return code


##
# ProjectFinder Implementations
##


class CurrentDirectoryProjectFinder(ProjectFinder):
    """
    Goes through a list of script finders and returns the first one matching.
    """

    def __init__(self, script_runners: Iterable[ScriptRunner]) -> None:
        self.script_runners = list(script_runners)

    def find_project(self, directory: Path) -> "ProjectInfo | None":
        for runner in self.script_runners:
            script = runner.find_script(directory)
            if script is not None:
                return ProjectInfo(script, runner)

        return None

    @classmethod
    def default(cls) -> "CurrentDirectoryProjectFinder":
        """
        Returns the default instance that contains the known :class:`ScriptRunner` implementations.
        """

        return cls(
            (
                PythonScriptRunner(),
                BuildDslScriptRunner(),
            )
        )


class GitAwareProjectFinder(ProjectFinder):
    """
    Finds the root of a project by picking the highest-up build script that does not cross a Git repository boundary
    or a home directory boundary. Starts from a directory and works it's way up until a stop condition is encountered.

    If any build script contains the string `# ::krakenw-root`, then the directory containing that build script is
    considered the root of the project. This is useful for projects that have multiple build scripts in different
    directories, but they should not be considered part of the same project.
    """

    def __init__(self, delegate: ProjectFinder, home_boundary: "Path | None | NotSet" = None) -> None:
        """
        :param delegate: The project finder to delegate to in any of the directories that this class
            looks through. The first project returned by this finder in any of the directories is used.
        :param home_boundary: A directory which contains a boundary that should not be crossed when
            searching for projects. For example, if this is `/home/foo`, then this class will stop searching
            for projects as soon as it reaches `/home/foo` or any sibling directory (such as `/home/bar`).
            If a path does not live within the home boundary or any of its siblings, the boundary is not
            taken into account.

            When the parameter is set to :const:`NotSet.Value`, it will default to the user's home directory.
            The value can be set to :const:`None` to disable the home boundary check all together.
        """
        self.delegate = delegate
        self.home_boundary = (
            Path("~").expanduser().parent.absolute() if home_boundary is NotSet.Value else home_boundary
        )

    def find_project(self, directory: Path) -> "ProjectInfo | None":
        highest_script: "ProjectInfo | None" = None
        directory = directory.absolute()
        while directory != Path(directory.root):
            # If we're in any directory that could be a home directory, we stop searching.
            if directory.parent == self.home_boundary:
                break

            script = self.delegate.find_project(directory)
            if script is not None:
                highest_script = script
                if script.script.read_text().find("# ::krakenw-root") != -1:
                    break

            # If in the next loop we would cross a Git repository boundary, we stop searching.
            if (directory / ".git").exists():
                break

            directory = directory.parent

        return highest_script

    @classmethod
    def default(cls) -> "GitAwareProjectFinder":
        """
        Returns the default instance that contains a default :class:`CurrentDirectoryProjectFinder`.
        """

        return cls(CurrentDirectoryProjectFinder.default())



from __future__ import annotations

import copy
import hashlib
import json
import logging
import os
import re
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Iterable, Iterator, Sequence

from typing_extensions import NotRequired, TypedDict

if TYPE_CHECKING:
    import rich.table

logger = logging.getLogger(__name__)
DEFAULT_CACHE_PATH = Path("~/.cache/krakenw/python-interpreters.json")


class InterpreterCandidate(TypedDict):
    path: str
    min_version: NotRequired[str | None]
    exact_version: NotRequired[str | None]


class Interpreter(TypedDict):
    path: str  # absolute path
    version: str  # x.y.z
    selected: NotRequired[bool]


class InterpreterVersionCache:
    """
    Helper class to cache the version of Python interpreters.
    """

    class _CacheEntry(TypedDict, total=True):
        md5sum: str
        version: str

    class _Payload(TypedDict, total=True):
        version: str
        paths: dict[str, InterpreterVersionCache._CacheEntry]

    EMPTY_PAYLOAD: ClassVar[_Payload] = {"version": "v1", "paths": {}}

    def __init__(self, file: Path = DEFAULT_CACHE_PATH) -> None:
        self.file = file.expanduser()
        self.cache: InterpreterVersionCache._Payload | None = None

    def _load(self) -> InterpreterVersionCache._Payload:
        if self.cache is None:
            if self.file.is_file():
                self.cache = json.loads(self.file.read_text())
            else:
                self.cache = copy.deepcopy(InterpreterVersionCache.EMPTY_PAYLOAD)
        return self.cache

    def _save(self) -> None:
        self.file.parent.mkdir(parents=True, exist_ok=True)
        self.file.write_text(json.dumps(self._load()))

    def _hash_file(self, path: Path) -> str:
        return hashlib.md5(path.read_bytes()).hexdigest()

    def get_version(self, path: Path) -> str | None:
        try:
            path = path.resolve()
        except FileNotFoundError:
            return None
        cache = self._load()
        entry = cache["paths"].get(str(path))
        if entry is None or entry["md5sum"] != self._hash_file(path):
            return None
        return entry["version"]

    def set_version(self, path: Path, version: str) -> None:
        path = path.resolve()
        cache = self._load()
        cache["paths"][str(path)] = {"md5sum": self._hash_file(path), "version": version}
        self._save()


def get_candidates(
    path_list: Sequence[str | Path] | None = None, check_pyenv: bool = True
) -> Iterator[InterpreterCandidate]:
    """
    Finds all Python interpreters on the system. This function only finds possible candidates, it does not
    check whether the interpreter is actually valid or retrieve its actual version. The generator may yield
    duplicate results if there are multiple symlinks pointing to the same interpreter binary.

    If *path_list* is not specified, the current PATH is used. If *check_pyenv* is True, Python interpreters
    installed via pyenv are also included in the results.
    """

    if path_list is None:
        path_list = os.environ["PATH"].split(os.pathsep)

    commands: set[Path] = set()
    for path in map(Path, path_list):
        if not path.is_dir():
            continue
        for item in path.iterdir():
            try:
                if not item.is_file() or not os.access(item, os.X_OK):
                    continue
            except PermissionError:
                continue
            if item.name.startswith("python") or item.name == "py":
                commands.add(item)

    # py and python
    for command in commands:
        if command.name in ("py", "python"):
            yield {"path": str(command)}

    # pythonX
    for command in commands:
        match = re.match(r"python(\d)$", command.name)
        if match:
            yield {
                "path": str(command),
                "min_version": f"{match.group(1)}.0.0" if match.group(1) else None,
            }

    # pythonX.Y
    for command in commands:
        match = re.match(r"python(\d\.\d\d?)$", command.name)
        if match:
            yield {"path": str(command), "min_version": f"{match.group(1)}.0"}

    # pythonX.Y.Z
    for command in commands:
        match = re.match(r"python(\d\.\d\d?\.\d\d?)$", command.name)
        if match:
            yield {"path": str(command), "exact_version": match.group(1)}

    # pyenv
    pyenv_versions = Path("~/.pyenv/versions").expanduser()
    if check_pyenv and pyenv_versions.is_dir():
        for item in pyenv_versions.iterdir():
            if re.match(r"\d+\.\d+\.\d+$", item.name) and item.is_dir():
                yield {"path": str(item / "bin" / "python"), "exact_version": item.name}

    yield {"path": sys.executable, "exact_version": ".".join(map(str, sys.version_info[:3]))}


def get_python_interpreter_version(python_bin: str) -> str:
    """
    Returns the version of the given Python interpreter by querying it with the `--version` option.

    Raises:
        RuntimeError: If the output of the command cannot be parsed.
        subprocess.CalledProcessError: If the command fails.
    """

    output = subprocess.check_output([python_bin, "--version"], stderr=subprocess.STDOUT, universal_newlines=True)
    match = re.match(r"Python (\d+\.\d+\.\d+)", output)
    if not match:
        raise RuntimeError(f"Could not determine Python version from output: {output}")
    return match.group(1)


def evaluate_candidates(
    candidates: Iterable[InterpreterCandidate], cache: InterpreterVersionCache | None = None
) -> list[Interpreter]:
    """
    Evaluates Python interpreter candidates and returns the deduplicated list of interpreters that were found.
    """

    interpreters: list[Interpreter] = []
    visited: set[Path] = set()

    for choice in candidates:
        try:
            path = Path(choice["path"]).resolve()
        except FileNotFoundError:
            logger.debug("Python interpreter %s does not exist", choice["path"], exc_info=True)
            continue

        if path in visited:
            continue
        visited.add(path)

        version = cache.get_version(path) if cache else None
        if version is None:
            try:
                version = get_python_interpreter_version(str(path))
            except (subprocess.CalledProcessError, RuntimeError):
                logger.debug("Failed to get version for Python interpreter %s", path, exc_info=True)
                continue
            if cache:
                cache.set_version(path, version)

        interpreter: Interpreter = {"path": str(path), "version": version}
        interpreters.append(interpreter)

    return interpreters


def build_rich_table(interpreters: Iterable[Interpreter]) -> rich.table.Table:
    """
    Gets a table of all viable Python interpreters on the system.

    Requires that the `rich` package is installed.
    """

    import rich.table

    tb = rich.table.Table("Path", "Version")
    for interpreter in interpreters:
        version = interpreter["version"]
        if interpreter.get("selected"):
            version += " *"
        tb.add_row(interpreter["path"], version)
    return tb


def match_version_constraint(constraint: str, version: str) -> bool:
    """
    Match a Python package version against a constraint.
    """

    from packaging.specifiers import SpecifierSet

    return SpecifierSet(constraint).contains(version)


def main() -> None:
    import argparse

    try:
        import rich
        import rich.table

        def tabulate(interpreters: Iterable[Interpreter]) -> None:
            rich.print(build_rich_table(interpreters))

    except ImportError:

        def tabulate(interpreters: Iterable[Interpreter]) -> None:
            for interpreter in interpreters:
                print(interpreter)

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--constraint", help="Find a Python interpreter with the given constraint.")
    args = parser.parse_args()

    interpreters = evaluate_candidates(get_candidates(), InterpreterVersionCache())

    if args.constraint:
        for interpreter in interpreters:
            if match_version_constraint(args.constraint, interpreter["version"]):
                interpreter["selected"] = True

    tabulate(interpreters)


if __name__ == "__main__":
    main()



import datetime

DATETIME_FORMATS = [
    "%Y-%m-%dT%H:%M:%S.%f%z",
    "%Y-%m-%dT%H:%M:%S.%f",
    # For backwards compatibility to read timestamps created with a previous version.
    "%Y-%m-%dT%H:%M:%S.Z",
]


def datetime_to_iso8601(dt: datetime.datetime) -> str:
    return dt.strftime(DATETIME_FORMATS[0])


def iso8601_to_datetime(value: str) -> datetime.datetime:
    err: "ValueError | None" = None
    for fmt in DATETIME_FORMATS:
        try:
            return datetime.datetime.strptime(value, fmt)
        except ValueError as exc:
            err = exc
    assert err is not None
    raise err



import sys
from pathlib import Path


def is_relative_to(apath: Path, bpath: Path) -> bool:
    """
    Checks if *apath* is a path relative to *bpath*.
    """

    if sys.version_info[:2] < (3, 9):
        try:
            apath.relative_to(bpath)
            return True
        except ValueError:
            return False
    else:
        return apath.is_relative_to(bpath)


def try_relative_to(apath: Path, bpath: "Path | None" = None) -> Path:
    """
    Tries to compute the relative path of *apath* relative to *bpath*. Returns the original *apath* if the
    relative path can be be computed, for exmaple if we would need to go at least one directory up to reach
    *apath* relative to *bpath*.
    """

    try:
        return apath.relative_to(bpath or Path.cwd())
    except ValueError:
        return apath



from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Iterator

from pytest import fixture

from kraken.common._buildscript import BuildscriptMetadata
from kraken.common._generic import not_none
from kraken.common._runner import (
    BuildDslScriptRunner,
    CurrentDirectoryProjectFinder,
    GitAwareProjectFinder,
    PythonScriptRunner,
)


@fixture
def tempdir() -> Iterator[Path]:
    with TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


def test__PythonScriptRunner__can_find_and_execute_script(tempdir: Path) -> None:
    (tempdir / ".kraken.py").write_text(
        dedent(
            """
        from kraken.common import buildscript

        buildscript(requirements=["kraken-std"])
        """
        )
    )
    runner = PythonScriptRunner()
    script = runner.find_script(tempdir)
    assert script is not None

    with BuildscriptMetadata.capture() as metadata_future:
        runner.execute_script(script, {})

    assert metadata_future.result() == BuildscriptMetadata(requirements=["kraken-std"])


def test__BuildDslScriptRunner__can_find_and_execute_script(tempdir: Path) -> None:
    (tempdir / "kraken.build").write_text(
        dedent(
            """
        buildscript {
            requires "kraken-std"
        }
        """
        )
    )

    runner = BuildDslScriptRunner()
    script = runner.find_script(tempdir)
    assert script is not None

    with BuildscriptMetadata.capture() as metadata_future:
        runner.execute_script(script, {})

    assert metadata_future.result() == BuildscriptMetadata(requirements=["kraken-std"])


def test__GitAwareProjectFinder__finds_highest_script(tempdir: Path) -> None:
    finder = GitAwareProjectFinder(CurrentDirectoryProjectFinder([PythonScriptRunner()]))

    (tempdir / "foo" / "bar").mkdir(parents=True)
    (tempdir / "foo" / ".kraken.py").write_text("")
    (tempdir / ".kraken.py").write_text("")  # <<< result

    script, _ = not_none(finder.find_project(directory=tempdir / "foo" / "bar"))
    assert script == tempdir / ".kraken.py"


def test__GitAwareProjectFinder__finds_highest_script_but_does_not_cross_home_boundary(tempdir: Path) -> None:
    finder = GitAwareProjectFinder(CurrentDirectoryProjectFinder([PythonScriptRunner()]), home_boundary=tempdir)

    # With the home_boundary set to tempdir, it will not pick scripts in any directory directly inside of tempdir,
    # e.g. tempdir/foo/.kraken.py is off-limits.

    (tempdir / "foo" / "bar").mkdir(parents=True)
    (tempdir / "foo" / "bar" / ".kraken.py").write_text("")  # <<< result
    (tempdir / "foo" / ".kraken.py").write_text("")
    (tempdir / ".kraken.py").write_text("")

    script, _ = not_none(finder.find_project(directory=tempdir / "foo" / "bar"))
    assert script == tempdir / "foo" / "bar" / ".kraken.py"


def test__GitAwareProjectFinder__finds_highest_script_but_does_not_cross_git(tempdir: Path) -> None:
    finder = GitAwareProjectFinder(CurrentDirectoryProjectFinder([PythonScriptRunner()]))

    (tempdir / "foo" / "bar").mkdir(parents=True)
    (tempdir / "foo" / "bar" / ".kraken.py").write_text("")
    (tempdir / "foo" / ".kraken.py").write_text("")  # <<< result
    (tempdir / "foo" / ".git").write_text("")
    (tempdir / ".kraken.py").write_text("")

    script, _ = not_none(finder.find_project(directory=tempdir / "foo" / "bar"))
    assert script == tempdir / "foo" / ".kraken.py"



from kraken.common._buildscript import BuildscriptMetadata, buildscript


def test_capture_kraken_buildscript() -> None:
    assert buildscript() == BuildscriptMetadata()
    assert buildscript(index_url="http://foo/simple") == BuildscriptMetadata(index_url="http://foo/simple")
    assert buildscript(requirements=["a"]) == BuildscriptMetadata(requirements=["a"])
    assert buildscript(extra_index_urls=["b"]) == BuildscriptMetadata(extra_index_urls=["b"])

    with BuildscriptMetadata.capture() as future:
        buildscript(index_url="c")

    assert future.result() == BuildscriptMetadata(index_url="c")



import enum
from typing import Mapping, MutableMapping

KRAKEN_ENVIRONMENT_TYPE_VARIABLE = "_KRAKEN_COMMONS_ENVIRONMENT_TYPE"


class EnvironmentType(enum.Enum):
    """
    This enumeration describes the type of environment that is being used to run Kraken in.
    """

    #: This enum reflects that Kraken was run directly, without being invoked through Kraken wrapper.
    NATIVE = 0

    #: Wrapper, using a virtual environment.
    VENV = 1

    #: Wrapper, using a PEX file.
    PEX_ZIPAPP = 2

    #: Wrapper, using a packed PEX environment.
    PEX_PACKED = 3

    #: Wrapper, using a loose PEX environment.
    PEX_LOOSE = 4

    def is_native(self) -> bool:
        return self == EnvironmentType.NATIVE

    def is_pex(self) -> bool:
        return self in (
            EnvironmentType.PEX_ZIPAPP,
            EnvironmentType.PEX_PACKED,
            EnvironmentType.PEX_LOOSE,
        )

    def is_venv(self) -> bool:
        return self == EnvironmentType.VENV

    def is_wrapped(self) -> bool:
        return self != EnvironmentType.NATIVE

    @staticmethod
    def get(environ: Mapping[str, str]) -> "EnvironmentType":
        value = environ.get(KRAKEN_ENVIRONMENT_TYPE_VARIABLE, EnvironmentType.NATIVE.name)
        try:
            return EnvironmentType(value)
        except ValueError:
            raise RuntimeError(
                f"The value of environment variable {KRAKEN_ENVIRONMENT_TYPE_VARIABLE}={value!r} is invalid, "
                f"valid values are {', '.join(x.name for x in EnvironmentType)}. The most likely cause of this "
                "error is that the version of kraken-wrapper and kraken-build are incompatible."
            )

    def set(self, environ: MutableMapping[str, str]) -> None:
        environ[KRAKEN_ENVIRONMENT_TYPE_VARIABLE] = self.name



import abc
import argparse
import dataclasses
import hashlib
import logging
import re
from pathlib import Path
from typing import Any, Dict, Iterable, List, Tuple

from ._buildscript import BuildscriptMetadata
from ._generic import NotSet, flatten

logger = logging.getLogger(__name__)
DEFAULT_BUILD_SUPPORT_FOLDER = "build-support"
DEFAULT_INTERPRETER_CONSTRAINT = ">=3.10,<3.11"  # Kraken-core requires 3.10 exactly


def parse_requirement(value: str) -> "PipRequirement | LocalRequirement":
    """
    Parse a string as a requirement. Return a :class:`PipRequirement` or :class:`LocalRequirement`.
    """

    match = re.match(r"(.+?)@(.+)", value)
    if match:
        return LocalRequirement(match.group(1).strip(), Path(match.group(2).strip()))

    match = re.match(r"([\w\d\-\_]+)(.*)", value)
    if match:
        return PipRequirement(match.group(1), match.group(2).strip() or None)

    raise ValueError(f"invalid requirement: {value!r}")


class Requirement(abc.ABC):
    name: str  #: The distribution name.

    @abc.abstractmethod
    def to_args(self, base_dir: Path) -> List[str]:
        """Convert the requirement to Pip args."""

        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class PipRequirement(Requirement):
    """Represents a Pip requriement."""

    name: str
    spec: "str | None"

    def __str__(self) -> str:
        return f"{self.name}{self.spec or ''}"

    def to_args(self, base_dir: Path) -> List[str]:
        return [str(self)]


@dataclasses.dataclass(frozen=True)
class LocalRequirement(Requirement):
    """Represents a requirement on a local project on the filesystem.

    The string format of a local requirement is `name@path`. The `name` must match the distribution name."""

    name: str
    path: Path

    def __str__(self) -> str:
        return f"{self.name}@{self.path}"

    def to_args(self, base_dir: Path) -> List[str]:
        return [str((base_dir / self.path if base_dir else self.path).absolute())]


@dataclasses.dataclass(frozen=True)
class RequirementSpec:
    """Represents the requirements for a kraken build script."""

    requirements: Tuple[Requirement, ...]
    index_url: "str | None" = None
    extra_index_urls: Tuple[str, ...] = ()
    interpreter_constraint: "str | None" = None
    pythonpath: Tuple[str, ...] = ()

    def __post_init__(self) -> None:
        for req in self.requirements:
            assert isinstance(req, Requirement), type(req)

    def __eq__(self, other: Any) -> bool:
        # NOTE (@NiklasRosenstein): packaging.requirements.Requirement is not properly equality comparable, so
        #       we implement a custom comparison based on the hash digest.
        if isinstance(other, RequirementSpec):
            return (type(self), self.to_hash()) == (type(other), other.to_hash())
        return False

    def with_requirements(self, reqs: Iterable["str | Requirement"]) -> "RequirementSpec":
        """Adds the given requirements and returns a new instance."""

        requirements = list(self.requirements)
        for req in reqs:
            if isinstance(req, str):
                req = parse_requirement(req)
            requirements.append(req)

        return self.replace(requirements=tuple(requirements))

    def with_pythonpath(self, path: Iterable[str]) -> "RequirementSpec":
        """Adds the given pythonpath and returns a new instance."""

        return self.replace(pythonpath=(*self.pythonpath, *path))

    def replace(
        self,
        requirements: "Iterable[Requirement] | None" = None,
        index_url: "str | None | NotSet" = NotSet.Value,
        extra_index_urls: "Iterable[str] | None" = None,
        interpreter_constraint: "str | None | NotSet" = NotSet.Value,
        pythonpath: "Iterable[str] | None" = None,
    ) -> "RequirementSpec":
        return RequirementSpec(
            requirements=self.requirements if requirements is None else tuple(requirements),
            index_url=self.index_url if index_url is NotSet.Value else index_url,
            extra_index_urls=self.extra_index_urls if extra_index_urls is None else tuple(extra_index_urls),
            interpreter_constraint=self.interpreter_constraint
            if interpreter_constraint is NotSet.Value
            else interpreter_constraint,
            pythonpath=self.pythonpath if pythonpath is None else tuple(pythonpath),
        )

    @staticmethod
    def from_json(data: Dict[str, Any]) -> "RequirementSpec":
        return RequirementSpec(
            requirements=tuple(parse_requirement(x) for x in data["requirements"]),
            index_url=data.get("index_url"),
            extra_index_urls=tuple(data.get("extra_index_urls", ())),
            interpreter_constraint=data.get("interpreter_constraint"),
            pythonpath=tuple(data.get("pythonpath", ())),
        )

    def to_json(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {"requirements": [str(x) for x in self.requirements], "pythonpath": self.pythonpath}
        if self.index_url is not None:
            result["index_url"] = self.index_url
        if self.extra_index_urls:
            result["extra_index_urls"] = self.extra_index_urls
        if self.interpreter_constraint:
            result["interpreter_constraint"] = self.interpreter_constraint
        return result

    @staticmethod
    def from_args(args: List[str]) -> "RequirementSpec":
        """Parses the arguments using :mod:`argparse` as if they are Pip install arguments.

        :raise ValueError: If an invalid argument is encountered."""

        parser = argparse.ArgumentParser()
        parser.add_argument("packages", nargs="*")
        parser.add_argument("--index-url")
        parser.add_argument("--extra-index-url", action="append")
        parser.add_argument("--interpreter-constraint")
        parsed, unknown = parser.parse_known_args(args)
        if unknown:
            raise ValueError(f"encountered unknown arguments in requirements: {unknown}")

        return RequirementSpec(
            requirements=tuple(parse_requirement(x) for x in parsed.packages or []),
            index_url=parsed.index_url,
            extra_index_urls=tuple(parsed.extra_index_url or ()),
            interpreter_constraint=parsed.interpreter_constraint,
        )

    def to_args(
        self,
        base_dir: Path = Path("."),
        with_options: bool = True,
        with_requirements: bool = True,
    ) -> List[str]:
        """Converts the requirements back to Pip install arguments.

        :param base_dir: The base directory that relative :class:`LocalRequirement`s should be considered relative to.
        :param with_requirements: Can be set to `False` to not return requirements in the argument, just the index URLs.
        """

        args = []
        if with_options and self.index_url:
            args += ["--index-url", self.index_url]
        if with_options:
            for url in self.extra_index_urls:
                args += ["--extra-index-url", url]
        if with_requirements:
            args += flatten(req.to_args(base_dir) for req in self.requirements)
        return args

    def to_hash(self, algorithm: str = "sha256") -> str:
        """Hash the requirements spec to a hexdigest."""

        hash_parts = [str(req) for req in self.requirements] + ["::pythonpath"] + list(self.pythonpath)
        hash_parts += ["::interpreter_constraint", self.interpreter_constraint or ""]
        return hashlib.new(algorithm, ":".join(hash_parts).encode()).hexdigest()

    @classmethod
    def from_metadata(cls, metadata: BuildscriptMetadata) -> "RequirementSpec":
        return RequirementSpec(
            requirements=tuple(map(parse_requirement, metadata.requirements)),
            index_url=metadata.index_url,
            extra_index_urls=tuple(metadata.extra_index_urls),
            interpreter_constraint=DEFAULT_INTERPRETER_CONSTRAINT,
            pythonpath=tuple(metadata.additional_sys_paths) + (DEFAULT_BUILD_SUPPORT_FOLDER,),
        )

    def to_metadata(self) -> BuildscriptMetadata:
        return BuildscriptMetadata(
            index_url=self.index_url,
            extra_index_urls=list(self.extra_index_urls),
            requirements=[str(x) for x in self.requirements],
            additional_sys_paths=[x for x in self.pythonpath if x != DEFAULT_BUILD_SUPPORT_FOLDER],
        )



from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Iterator

import tomli
from pytest import fixture

from kraken.common._tomlconfig import TomlConfigFile


@fixture
def tempdir() -> Iterator[Path]:
    with TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


def test__TomlConfigFile__load_and_save(tempdir: Path) -> None:
    config_file = tempdir / "config.toml"
    config_file.write_text(
        dedent(
            """
        [section]
        key = "value"
        """
        )
    )
    config = TomlConfigFile(config_file)
    assert dict(config) == {"section": {"key": "value"}}
    config["foo"] = "bar"
    config.save()

    assert tomli.loads(config_file.read_text()) == {"section": {"key": "value"}, "foo": "bar"}



import logging
import sys
from typing import Any, Callable, Type, TypeVar

logger = logging.getLogger(__name__)


T_Callable = TypeVar("T_Callable", bound=Callable[..., Any])


def exit_on_known_exceptions(
    *exception_types: Type[BaseException], log: bool = True, exit_code: int = 1
) -> Callable[[T_Callable], T_Callable]:
    """
    A useful decorator for CLI entrypoints that catches known exceptions and exits with a non-zero exit code.
    """

    def decorator(func: T_Callable) -> T_Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except exception_types:
                if log:
                    logger.debug("Exiting due to known exception", exc_info=True)
                sys.exit(exit_code)

        return wrapper  # type: ignore[return-value]

    return decorator



__version__ = "0.30.2"

from . import path
from ._argparse import propagate_argparse_formatter_to_subparser
from ._asciitable import AsciiTable
from ._buildscript import BuildscriptMetadata, BuildscriptMetadataException, buildscript
from ._date import datetime_to_iso8601, iso8601_to_datetime
from ._environment import EnvironmentType
from ._fs import atomic_file_swap, safe_rmpath
from ._generic import NotSet, flatten, not_none
from ._importlib import appending_to_sys_path, import_class
from ._option_sets import LoggingOptions
from ._requirements import LocalRequirement, PipRequirement, Requirement, RequirementSpec, parse_requirement
from ._runner import (
    BuildDslScriptRunner,
    CurrentDirectoryProjectFinder,
    GitAwareProjectFinder,
    ProjectFinder,
    ProjectInfo,
    PythonScriptRunner,
    ScriptPicker,
    ScriptRunner,
)
from ._terminal import get_terminal_width
from ._text import inline_text, lazy_str, pluralize
from ._tomlconfig import TomlConfigFile
from .supplier import Supplier

__all__ = [
    # _argparse
    "propagate_argparse_formatter_to_subparser",
    # _asciitable
    "AsciiTable",
    # _date
    "datetime_to_iso8601",
    "iso8601_to_datetime",
    # _environment
    "EnvironmentType",
    # _fs
    "atomic_file_swap",
    "safe_rmpath",
    # _generic
    "flatten",
    "not_none",
    "NotSet",
    # _importlib
    "import_class",
    "appending_to_sys_path",
    # _buildscript
    "buildscript",
    "BuildscriptMetadata",
    "BuildscriptMetadataException",
    # _option_sets
    "LoggingOptions",
    # _requirements
    "parse_requirement",
    "Requirement",
    "LocalRequirement",
    "PipRequirement",
    "RequirementSpec",
    # _runner
    "BuildDslScriptRunner",
    "CurrentDirectoryProjectFinder",
    "GitAwareProjectFinder",
    "ProjectFinder",
    "ProjectInfo",
    "PythonScriptRunner",
    "ScriptPicker",
    "ScriptRunner",
    # _terminal
    "get_terminal_width",
    # _text
    "pluralize",
    "inline_text",
    "lazy_str",
    # _tomlconfig
    "TomlConfigFile",
    # global
    "path",
    # supplier
    "Supplier",
]



from __future__ import annotations


def as_bytes(v: str | bytes, encoding: str) -> bytes:
    return v.encode(encoding) if isinstance(v, str) else v


def as_string(v: str | bytes, encoding: str) -> str:
    return v.decode(encoding) if isinstance(v, bytes) else v



from pathlib import Path
from typing import Any, Dict, Iterator, MutableMapping

import tomli
import tomli_w


class TomlConfigFile(MutableMapping[str, Any]):
    """
    A helper class that reads and writes a TOML configuration file.
    """

    def __init__(self, path: Path) -> None:
        self.path = path
        self._data: "Dict[str, Any] | None" = None

    def _get_data(self) -> "Dict[str, Any]":
        if self._data is None:
            if self.path.is_file():
                self._data = tomli.loads(self.path.read_text())
            else:
                self._data = {}
        return self._data

    def __getitem__(self, key: str) -> Any:
        return self._get_data()[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._get_data()[key] = value

    def __delitem__(self, key: str) -> None:
        del self._get_data()[key]

    def __len__(self) -> int:
        return len(self._get_data())

    def __iter__(self) -> Iterator[str]:
        return iter(self._get_data())

    def save(self) -> None:
        self.path.parent.mkdir(exist_ok=True, parents=True)
        self.path.write_text(tomli_w.dumps(self._get_data()))
